// SillyPhr.h
// Silly Phrase Generator, a.k.a. (the first?)
//  non-linear documentation language.

#ifndef SILLYPHR_H
#define SILLYPHR_H

class CSillyPhrase : public CObject
{
  public:
		CSillyPhrase();
		CSillyPhrase(const CString &fileName);
		~CSillyPhrase();

		// loads up the Silly Script fileName
		//  return value is whether the load was successfell
		bool loadDefinition(const CString &fileName);
		// spits out a silly phrase, by running the pharser on
		// the data loaded by loadDefinition.
		//  seed is the seed for the random number generator,
		//   it can be used to reproduce the same phrase.
		//   0 makes it seed with the system timer
		//  impressive controls the likely hood of optional replacements 
		CString makeAPhrase(UINT seed = 0, int impressive = -1);
		// dubuging function, returns the entire rule database
		CString getAll(void);

		// adds a warning to the warning list
		void addWarning(CString &aWarning);
		// adds an error to the error list
		void addError(CString &anError);
		// returns the warning list
		CString getWarnings(void);
		// returns the error list
		CString getErrors(void);
		// deletes the warning list
		void clearWarnings(void);
		// deletes the error list
		void clearErrors(void);

	protected:
		//data
		// load definition sets this as whether the last
		// load attempt succeded, it will have the same
		// value as it's return value.
		bool isLoaded;
		// the rule database
		CObArray definition;
		// the warning and error lists
		CStringArray warnings, errors;
		// the current maximum finished silly phrase length
		int maxLength;

		//loader trans-function variables
		// the current file being loaded.
		CStdioFile l_InputStream;
		// the string currently be worked on, 
		// and the file name currently being worked on
		CString l_InputString, l_FileName;
		// the number of the line currently being worked on
		//   (aka l_InputString)
		int l_LineNumber;

		//functions
		// the postProcessor, takes the almost finished phrase
		//  as an argument, and strips "s from it
		void postProcess(CString &thePhrase);
		// takes the rule name in toBeReplaced, finds it's
		//  definition, chooses a random value from that,
		//  and sets toBeReplaced to the value
		void getReplacement(CString &toBeReplaced);
		// searches the rule database for the rule which,
		//  and returns a pointer to it.
		CStringArray* getRule(const CString &which);
		// clears out the rule database
		void deleteAllStrings(void);

		//loader helpers
		// returns the next character from the input stream,
		//  lazily loading as neccisary and return a few
		//  special values
		unsigned char fetch(void);
		// puts a character back on input stream (l_InputString)
		void shove(unsigned char what);
		// decodes the file exception number, and turns
		//  into an error string
		void reportException(CFileException *e);
};

#endif