// SillyPhr.cpp

#include "stdafx.h"
#include "SillyPhr.h"
#include "..\\..\\common\\FIOErr.h"

// the system name for where the pharser starts working
const char MAINPHRASE[11] = "[**MAIN**]";

CSillyPhrase::CSillyPhrase(void)
	: isLoaded(false)
{}

CSillyPhrase::CSillyPhrase(const CString &fileName)
	: isLoaded(false)
{
  loadDefinition(fileName);
}

CSillyPhrase::~CSillyPhrase(void)
{
	deleteAllStrings();
}

bool CSillyPhrase::loadDefinition(const CString &fileName)
{
	// strings in progress, and the string used as a temporary for
	//  reporting erros and warnings
  CString virtualLine, keywordInProgress, foo;
	// pointers to the current and last rules
	CStringArray* currentRule = NULL, *lastRule = NULL;
	// defines the translation modes
	enum transMode {TMLINE, TMMULTILINE, TMFORMAT};
	// the current translation mode
	transMode translationMode = TMLINE;
	// the character being examined at the moment
	unsigned char testing;
	// special mode flags
	bool commentMode = false, defineMode = false;
	bool setMaxLengthMode = false, literalMode = false;
	// bracket depth counters
	int round = 0, pointy = 0;
	int square = 0, curly = 0;
	// used to retrieve any errors from the open call
	CFileException e;

	// setup
	l_LineNumber = 0;
	l_InputString.Empty();

	clearWarnings();
	clearErrors();

	// delete the old database
	if (isLoaded)
		deleteAllStrings();

	//set up system defined rules
	currentRule = new CStringArray;
	currentRule->Add(MAINPHRASE);
	definition.Add(currentRule);

	currentRule = new CStringArray;
	currentRule->Add("[cr]");
	currentRule->Add("\r\n");
	definition.Add(currentRule);

	currentRule = new CStringArray;
	currentRule->Add("[tab]");
	currentRule->Add("\t");
	definition.Add(currentRule);

	// if no rule name is specified at the start, items
	//  got into the main rule
	lastRule = currentRule = getRule(MAINPHRASE);

	// default maximum length
	maxLength = 256;

	// try to open the file....
	if (l_InputStream.Open(fileName, CFile::modeRead, &e))
	{
		if (fetch() == 254)  // pump the character grabber to get the first line in.
			return (isLoaded = false);  // but check for empty files.
		// character loop
		while ((testing = fetch()) != 255)
		{
			#ifdef _DEBUG
			if (testing == '!')
				maxLength = 1000; // put a breakpoint here
													//  to break in at trouble spots
			#endif
			// skip everthing else if in comment mode
			if (commentMode)
			{
				// check for end of comment
				if (testing == '\0')
				{
					commentMode = false;
					// shove the end-of-line back on so it can be
					//  handled properly for the translation mode
					shove('\0');
				}
			}
			// check for start/end of literal mode
			else if (testing == '"')
			{
				literalMode = !literalMode;
				virtualLine += '"';
			}
			// if it's an end-of-line....
			else if (testing == '\0')
			{
				// if getting the number for the maximum size, 
				//  end the line and process it later
				//  Make sure we arn't in literal mode any more.
				if (setMaxLengthMode)
				{
					shove('|');
					literalMode = false;
				}
				// otherwise, the action depends on the translation mode
				else
				{
					switch (translationMode)
					{
						// line mode, break item
						case TMLINE:
							shove('|');
							literalMode = false;
							break;
						// multiline mode, check for a blank line,
						//  and if so, break item
						case TMMULTILINE:
							switch (testing = fetch())
							{
								case 254:
									shove(255);
									shove('|');
									break;
								case '\0':
									shove('|');
									literalMode = false;
									break;
								default:
									shove(testing);
									break;
							}
							break;
						// format mode, convert it to a carrige return
						case TMFORMAT:
							virtualLine += "\r\n";
							break;
					}
				}
			}
			// it's a ~, eat a /0 (end-of-line)
			else if (testing == '~')
			{
				// eats a following real-line break
				if ((testing = fetch()) != '\0')
				{
				  shove(testing);
					virtualLine += '~';
				}
			}
			// end-of-file detected
			else if (testing == 254)
			{
				// break entry and end the loop
				shove(255);
				shove('|');
				// who forgot to close thier "s ?
				if (literalMode)
				{
					foo.Format("unterminated literal quote at end of file");
					addWarning(foo);
					literalMode = false;
				}
			}
			// as long as it's not literal mode, check everything else
			else if (!literalMode)
			{
				switch (testing)
				{
					// next couple keep nesting counts and warn
					//  on any mismatched pairs.
					case '<': 
						pointy++; 
						virtualLine += '<';
						break;
					case '>':
						pointy--;
						if (pointy < 0)
						{
							foo.Format("line %d has mis-matched <>", 
								l_LineNumber);
							addWarning(foo);
						}
						virtualLine += '>';
						break;
					case '(': 
						round++; 
						virtualLine += '(';
						break;
					case ')':
						round--;
						if (round < 0)
						{
							foo.Format("line %d has mis-matched ()", 
								l_LineNumber);
							addWarning(foo);
						}
						virtualLine += ')';
						break;
					case '/':
						commentMode = true;
						break;
					case '[': 
						square++; 
						if (square > 1)
						{
							foo.Format("ine %d: [ does not have matching ]", 
								l_LineNumber);
							addError(foo);
						}
						if (defineMode)
							keywordInProgress += '[';
						else
							virtualLine += '[';
						break;
					case ']':
						square--;
						if (square < 0)
						{
							foo.Format("line %d ] does not have matching [", 
								l_LineNumber);
							addError(foo);
						}
						// if the rule name just closed was after a define....
						else if (defineMode)
						{
							keywordInProgress += ']';
							// try to find the rule....
							currentRule = getRule(keywordInProgress);
							// if not, create it.
							if (currentRule == NULL)
							{
								currentRule = new CStringArray;
								currentRule->Add(keywordInProgress);
								definition.Add(currentRule);
							}
							// end define mode and break item to clear the buffer
							//  into the last rule
							defineMode = false;
							keywordInProgress.Empty();
							shove('|');
						}
						else
							virtualLine += ']';
						break;// end case ']'
					case '{':
						curly++;
						if (curly > 1)
						{
							foo.Format("line %d: { does not have matching }", 
								l_LineNumber);
							addError(foo);
						}
						keywordInProgress += '{';
						break;
					case '}':
						curly--;
						if (curly < 0)
						{
							foo.Format("line %d } does not have matching {", 
								l_LineNumber);
							addError(foo);
						}
						keywordInProgress += '}';
						// a keyword just ended, decide which one
						if (keywordInProgress == "{Define}")
						{
							// start looking for a rule name
							defineMode = true;
						}
						else if (keywordInProgress == "{Main}")
						{
							// set it to the main rule, and clear the buffer.
							currentRule = getRule(MAINPHRASE);
							shove('|');
						}
						// translation modes
						else if (keywordInProgress == "{Line}")
							translationMode = TMLINE;
						else if (keywordInProgress == "{Multiline}")
							translationMode = TMMULTILINE;
						else if (keywordInProgress == "{Format}")
							translationMode = TMFORMAT;
						// start looking for a number
						else if (keywordInProgress == "{SetMaxLength}")
							setMaxLengthMode = true;
						// oops...
						else
						{
							foo.Format("line %d: unrecongized keyword", 
								l_LineNumber);
							addError(foo);
						}
						keywordInProgress.Empty();
						break; // end case '}'
					case '|':
						// if we are still in define mode, somebody forgot
						//  to give us a rule name to define.
						if (defineMode)
						{
							foo.Format("line %d: {Define} missing rule name", 
								l_LineNumber);
							addError(foo);
							defineMode = false;
						}
						// there should be a number in keywordInProgress
						else if (setMaxLengthMode)
						{
							maxLength = atoi(keywordInProgress);
							if (maxLength < 1)
							{
								foo.Format("line %d sets an illegal maximum length: %d", 
									l_LineNumber, maxLength);
								addError(foo);
							}
							else if (maxLength < 32)
							{
								foo.Format("line %d sets a suspicouly low maximum length: %d", 
									l_LineNumber,  maxLength);
								addWarning(foo);
							}
							keywordInProgress.Empty();
							setMaxLengthMode = false;
						}
						// otherwise, as long as the line is progress is not empty
						//  add it to the current rule	
						else
						{
							if (!virtualLine.IsEmpty())
							{
								// lastRule and currentRule are used so the {Define}
								//  flushes go to the right place
								lastRule->Add(virtualLine);
								virtualLine.Empty();
							}
						}
						lastRule = currentRule;
						// make sure all the brackets are in order
						if (round != 0)
						{
							foo.Format("Item ending on line %d (or the line before it) has mis-matched ()", 
								l_LineNumber);
							addWarning(foo);
							round = 0;
						}
						else if (pointy != 0)
						{
							foo.Format("Item ending on line %d (or the line before it) has mis-matched <>", 
								l_LineNumber);
							addWarning(foo);
							pointy = 0;
						}
						else if (square != 0)
						{
							foo.Format("Item ending on line %d (or the line before it) has mis-matcched []", 
								l_LineNumber);
							addWarning(foo);
							square = 0;
						}
						else if (curly != 0)
						{
							foo.Format("Item ending on line %d (or the line before it) has mis-matcched {}", 
								l_LineNumber);
							addWarning(foo);
							curly = 0;
						}
						break; // end case '|'
					default:
						// if a { has not yet been matched by a },
						//  we must be makeing a keyword
						if (curly == 1)
						{
						  keywordInProgress += testing;
						}
						// if after a {Define}, and inside [], must be
						//  rule name in progress
						else if (defineMode)
						{
							if (square == 1)
								keywordInProgress += testing;
							// else it goes to the bitbucket
						}
						// if after a {SetMaxLength}, it better be a number
						else if (setMaxLengthMode)
						{
							if (testing >= '0' && testing <= '9')
							  keywordInProgress += testing;
							//else throw it away.
						}
						// otherwise chuck it on the main work buffer
						else
						{
						  virtualLine += testing;
						}
						break; // end default
				}// end switch on character
			} // end if !literalmode
			// if all else fails, chuck it on the main work buffer
			else
				virtualLine += testing;	
		}// end character loop
	} // end if file opened OK
	// if it didn't open ok, complain
	else
	{
		reportException(&e);
		return (isLoaded = false);
	}
	// make sure the file closes ok
	try
	{
		l_InputStream.Close();
	}
	catch (CFileException *e)
	{
		reportException(e);
	}
	// return whether there were errors or not
	if (errors.GetSize() < 1)
		return (isLoaded = true);
	else
		return (isLoaded = false);
}

unsigned char CSillyPhrase::fetch(void)
{
	if (l_InputString.IsEmpty())
	{
		// time to get a new line
		l_LineNumber++;
		// return end-of-line if we got a new one,
		//  else return end-of-file
		if (l_InputStream.ReadString(l_InputString))
		  return '\0';
		else
			return (unsigned char)254;
	}
	// otherwise just return the next character
	else
	{
		char returnValue = l_InputString[0];
		l_InputString = l_InputString.Right(l_InputString.GetLength() - 1);
		return returnValue;
	}
}

void CSillyPhrase::shove(unsigned char what)
{
	l_InputString = (char)what + l_InputString;
}

void CSillyPhrase::reportException(CFileException *e)
{
	addError(decodeFileException(e) + ": " + l_FileName);
}

CString CSillyPhrase::getAll(void)
{
  CStringArray* aStringArray;
	CString outString;
	int a, b;

	// for each rule
  for (a = definition.GetSize()-1;a >= 0;a--)
	{
		// and each item in the rule (which includes the rule name as #0
	  aStringArray = (CStringArray*)definition[a];
		for (b = aStringArray->GetSize()-1;b >= 0;b--)
		{
			// stick on the output string
		  outString = outString + aStringArray->GetAt(b) + "/";
		}
		// and make it readable
		outString += "\r\n";
	}
	return outString;
}


CString CSillyPhrase::makeAPhrase(UINT seed, int impressive)
{
	// make sure we have something legitimate to work with
  if (!isLoaded)
	  return getErrors();

	// handle the special seed case
  if (seed == 0)
		srand( (unsigned) time(NULL));
	else
    srand(seed);

	// curb impressive so we don't get infinte optional-repeating sections
	if (impressive > 9)
	  impressive = 9;

	// the percent chance of optional stuff getting stuck in
	int percent = RAND_MAX * impressive / 10;
	// the main work buffer, and a temp
	CString phrase, temp;
	// tags for the opening and closing brackets in a pair
	int tagOpen = 0, tagClose;
	// the opeing bracket, it's matching closing bracket,
	//  and the character we are currently looking at.
	char doing, closes, testing;
	// mode controls
	bool running = true, literalMode = false;
	// the index of testing, a place to keep phrase's length
	//   and avoid function calls, and the nesting depth
	//   of same type brackets
	int testIndex, length, nesting = 0;

	// make the psuedo-random number generator seem a little more random
	rand();
	rand();
	rand();
	rand();

	// start with the systems main phrase
	phrase = MAINPHRASE;

	// loop as long as we have something to do 
	//  and we don't exceede the maximum length
	while (running && ((length = phrase.GetLength()) < maxLength) )
	{
		// asume nothing left
		running = false;
		closes = 0;
		// look for something to do
		for (testIndex = tagOpen;testIndex < length;testIndex++)
		{
		  testing = phrase[testIndex];
			// if inside "s, check for the end or ignore the rest of the loop
			if (literalMode)
			{
			  if (testing == '\"')
				  literalMode = false;
				continue;
			}

			// don't bother if we have found and opening bracket already
			if (closes == 0)
			{
				switch(testing)
				{
				  case '(': 
						tagOpen = testIndex;
						doing = testing; 
						closes = ')'; 
						break;
					case '<': 
						tagOpen = testIndex; 
						doing = testing;
						closes = '>'; 
						break;
					case '[': 
						tagOpen = testIndex; 
						doing = testing;
						closes = ']'; 
						break;
					case '"': literalMode = true; break;
				}
			}
			// if we have, look for the matching closing brace
			else if (testing == closes)
			{
				// take care of nesting, only worry about a bracket
				//  at the same level
				if (--nesting < 1)
				{
					tagClose = testIndex;
					running = true;
					break;
				}
			}

			// if a same bracket is found, take note
			if (testing == doing)
			{
				nesting++;
			}
		}
		//  stop!
		if (!running)
		{
			break;
		}
		// do all that bracket specific stuff
		switch (doing)
		{
		  case '(':
				// cut out the stuff between the ()
				temp = phrase.Mid(tagOpen + 1, tagClose - tagOpen - 1);
				if (rand() < percent)
					// if so, just put it back together with the ()
					phrase = phrase.Left(tagOpen) 
						+ temp
						+	phrase.Right(phrase.GetLength() - tagClose - 1);
				else
					// otherwise, just put the ends together without the middle
					phrase = phrase.Left(tagOpen) 
						+	phrase.Right(phrase.GetLength() - tagClose - 1);
				break;
			case '<':
				// cut the stuff between the <>
				temp = phrase.Mid(tagOpen + 1, tagClose - tagOpen - 1);
				if (rand() < percent)
					// if so, put it before the <> 
					phrase = phrase.Left(tagOpen) 
						+ temp
						+	phrase.Right(phrase.GetLength() - tagOpen);
				else
					// if not, paste the end together
					phrase = phrase.Left(tagOpen) 
						+	phrase.Right(phrase.GetLength() - tagClose - 1);
				break;
			case '[':
				// cut out the rule name, including []
				temp = phrase.Mid(tagOpen, tagClose - tagOpen + 1);
				// choose a replacement for it
				getReplacement(temp);
				// and stick it back where the rule name was
				phrase = phrase.Left(tagOpen) 
					+ temp
					+	phrase.Right(phrase.GetLength() - tagClose - 1);
				break;
		}
	}
	// run the post-processor to strip out "s
	postProcess(phrase);
	return phrase;
}

//it's own seperate function for clarity
void CSillyPhrase::postProcess(CString &thePhrase)
{
	int testIndex, length;

	// Check to see if the maximum length was exceeded,
	//  and if so, stick an error reprot string in.
	// Do this before stripping quotes because that could
	//  reduce the size
	CString tackString = "!!Maximum length exceeded!!";
	if (thePhrase.GetLength() > maxLength)
		thePhrase = thePhrase.Left(maxLength - tackString.GetLength()) + tackString;

	// remember the length to save function calls
	length = thePhrase.GetLength();

	for (testIndex = 0;testIndex < length;testIndex++)
	{
		if (thePhrase[testIndex] == '\"')
		{
			// Whatever character comes immediatly after the one removed
			//  isn't even tested.  This what allows "" to become a "
			thePhrase = thePhrase.Left(testIndex) + thePhrase.Right(length - testIndex - 1);
			length = thePhrase.GetLength();
		}
	}

}


void CSillyPhrase::addWarning(CString &aWarning)
{
	warnings.Add("--Warning: " + aWarning);
}

void CSillyPhrase::addError(CString &anError)
{
	errors.Add("--Error: " + anError);
}

CString CSillyPhrase::getWarnings(void)
{
	CString compilation;
	compilation.Format("There where %d warnings on the last load\r\n", warnings.GetSize());
  for (int a = warnings.GetSize()-1;a >= 0;a--)
	  compilation += warnings[a] + "\r\n";
	return compilation;
}

CString CSillyPhrase::getErrors(void)
{
	CString compilation;
	compilation.Format("There where %d errors on the last load\r\n", errors.GetSize());
  for (int a = errors.GetSize()-1;a >= 0;a--)
	  compilation += errors[a] + "\r\n";
	return compilation;
}

void CSillyPhrase::clearWarnings(void)
{
	warnings.RemoveAll();
}

void CSillyPhrase::clearErrors(void)
{
	errors.RemoveAll();
}

void CSillyPhrase::getReplacement(CString &toBeReplaced)
{
	// I called the groups before I called the rules
	CStringArray* aGroup;

	// find the rule that goes by this name (toBeReplaced)
	aGroup = getRule(toBeReplaced);
	// if a rule wasn't found, report everbody's favorite
	//  in-line pharser error
	if (aGroup == NULL)
		toBeReplaced = "!!Replace Error!!";
	else
	{	
		// otherwise, some special case handling needs to be done
		switch (aGroup->GetSize())
		{
			// zero item rule
			case 1: toBeReplaced = ""; break;
			// one itme rule
			case 2: toBeReplaced = aGroup->GetAt(1); break;
			// everything else: enie, menie, miny, moe....
			default: 
				toBeReplaced = aGroup->GetAt(rand() % (aGroup->GetSize()-1) + 1); 
				break;
		}
	}
}

CStringArray* CSillyPhrase::getRule(const CString &which)
{
	 int a;
	CStringArray* aRule;

	// pretty simple; compare vs all the rule names,
	//  return the match if found, or NULL if not.
	for (a = definition.GetSize()-1;a >= 0;a--)
	{
	  aRule = (CStringArray*)definition[a];
		if (aRule->GetAt(0) == which)
			return aRule;
	}
	return NULL;
}

void CSillyPhrase::deleteAllStrings(void)
{
	// cleans out the rule database
  CObject *aPointer;
  while (definition.GetSize() > 0)
	{
	  aPointer = definition[0];
		definition.RemoveAt(0);
		delete aPointer;
	}
}

